#version 450

#if defined(VULKAN) || defined(GL_core_profile) || defined(GL_ES)

	#define int2 ivec2
	#define int4 ivec4
	#define uint2 uvec2
	#define uint4 uvec4
	#define float2 vec2
	#define float3 vec3
	#define float4 vec4
	#define lerp mix
	#define frac fract
	#define asint floatBitsToInt
	#define asuint floatBitsToUint
	#define asfloat uintBitsToFloat
	#define saturate(x) clamp(x, 0.0, 1.0)

	#if defined(GL_ES)

		precision highp float;
		precision highp int;
		precision highp sampler2D;
		precision highp usampler2D;

	#endif

#elif defined(__METAL_VERSION__)

	#include <metal_math>
	#include <metal_common>
	#include <metal_integer>
	#include <metal_graphics>
	#include <metal_texture>

	using namespace metal;

	#define lerp mix
	#define frac fract
	#define asint as_type<int>
	#define asuint as_type<uint>
	#define asfloat as_type<float>

#endif

#define kLogBandTextureWidth 12

#if defined(VULKAN) || defined(GL_core_profile) || defined(GL_ES)

	#define CurveSampler sampler2D
	#define BandSampler usampler2D
	#define TexelLoad2D(x, y) texelFetch(x, y, 0)

#elif defined(__METAL_VERSION__)

	#define CurveSampler texture2d<float, access::read>
	#define BandSampler texture2d<uint, access::read>
	#define TexelLoad2D(x, y) x.read(uint2(y))

#else

	#define CurveSampler Texture2D
	#define BandSampler Texture2D<uint4>
	#define TexelLoad2D(x, y) x.Load(int3(y, 0))

#endif

#if defined(SLUG_LOP3) || defined(GL_NV_gpu_shader5)

	#define slug_code int2

	int2 CalcRootCode(float y1, float y2, float y3)
	{
		int a = asint(y1);
		int b = asint(y2);
		int c = asint(y3);
		return (int2(~a & (b | c) | (~b & c), a & (~b | ~c) | (b & ~c)));
	}

	bool TestCurve(int2 code) {return ((code.x | code.y) < 0);}
	bool TestRoot1(int2 code) {return (code.x < 0);}
	bool TestRoot2(int2 code) {return (code.y < 0);}

#else

	#define slug_code uint

	uint CalcRootCode(float y1, float y2, float y3)
	{
		uint i1 = asuint(y1) >> 31U;
		uint i2 = asuint(y2) >> 30U;
		uint i3 = asuint(y3) >> 29U;

		#if defined(__PSSL__)

			uint shift = BitFieldInsert(2U, i2, i1);
			shift = BitFieldInsert(4U, i3, shift);

		#else

			uint shift = (i2 & 2U) | (i1 & ~2U);
			shift = (i3 & 4U) | (shift & ~4U);

		#endif

		return ((0x2E74U >> shift) & 0x0101U);
	}

	bool TestCurve(uint code) {return (code != 0U);}
	bool TestRoot1(uint code) {return ((code & 1U) != 0U);}
	bool TestRoot2(uint code) {return (code > 1U);}

#endif

float2 SolveHorizPoly(float4 p12, float2 p3)
{
	float2 a = p12.xy - p12.zw * 2.0 + p3;
	float2 b = p12.xy - p12.zw;
	float ra = 1.0 / a.y;
	float rb = 0.5 / b.y;

	float d = sqrt(max(b.y * b.y - a.y * p12.y, 0.0));
	float t1 = (b.y - d) * ra;
	float t2 = (b.y + d) * ra;

	if (abs(a.y) < 1.0 / 65536.0) t1 = t2 = p12.y * rb;

	return (float2((a.x * t1 - b.x * 2.0) * t1 + p12.x, (a.x * t2 - b.x * 2.0) * t2 + p12.x));
}

float2 SolveVertPoly(float4 p12, float2 p3)
{
	float2 a = p12.xy - p12.zw * 2.0 + p3;
	float2 b = p12.xy - p12.zw;
	float ra = 1.0 / a.x;
	float rb = 0.5 / b.x;

	float d = sqrt(max(b.x * b.x - a.x * p12.x, 0.0));
	float t1 = (b.x - d) * ra;
	float t2 = (b.x + d) * ra;

	if (abs(a.x) < 1.0 / 65536.0) t1 = t2 = p12.x * rb;

	return (float2((a.y * t1 - b.y * 2.0) * t1 + p12.y, (a.y * t2 - b.y * 2.0) * t2 + p12.y));
}

int2 CalcBandLoc(int2 glyphLoc, uint offset)
{
	int2 bandLoc = int2(glyphLoc.x + int(offset), glyphLoc.y);
	bandLoc.y += bandLoc.x >> kLogBandTextureWidth;
	bandLoc.x &= (1 << kLogBandTextureWidth) - 1;
	return (bandLoc);
}

float CalcCoverage(float xcov, float ycov, float xwgt, float ywgt, int flags)
{
	float coverage = max(abs(xcov * xwgt + ycov * ywgt) / max(xwgt + ywgt, 1.0 / 65536.0), min(abs(xcov), abs(ycov)));

	#if defined(SLUG_EVENODD)

		if ((flags & 0x1000) == 0)
		{

	#endif

			coverage = saturate(coverage);

	#if defined(SLUG_EVENODD)

		}
		else
		{
			coverage = 1.0 - abs(1.0 - frac(coverage * 0.5) * 2.0);
		}

	#endif

	#if defined(SLUG_WEIGHT)

		coverage = sqrt(coverage);

	#endif

	return (coverage);
}

#if defined(SLUG_MULTICOLOR)

	float4 ApplyLayerColor(uint2 layerData, float coverage, float4 finalColor)
	{
		#if defined(__PSSL__)

			float3 layerColor = float3(UnpackByte0(layerData.x), UnpackByte1(layerData.x), UnpackByte0(layerData.y)) * (1.0 / 255.0);

		#else

			float3 layerColor = float3(float(layerData.x & 0xFFU), float(layerData.x >> 8U), float(layerData.y & 0xFFU)) * (1.0 / 255.0);

		#endif

		return (float4(lerp(finalColor.xyz, layerColor, coverage), saturate(finalColor.w + coverage)));
	}

#endif

float4 RenderGlyph(CurveSampler curveData, BandSampler bandData, float2 pixelPosition, float4 vertexColor, float4 bandTransform, int4 glyphData)
{
	int curveIndex;

	float2 renderCoord = pixelPosition;
	float2 emsPerPixel = fwidth(renderCoord);
	float2 pixelsPerEm = 1.0 / emsPerPixel;

	int2 bandMax = glyphData.zw;
	bandMax.y &= 0x00FF;

	#if defined(SLUG_SUPER)

		int sampleIndex;
		int2 bandIndex;

		int2 sampleCount = clamp(int2(emsPerPixel * 16.0 + 1.0), int2(1, 1), int2(4, 4));

	#else

		int2 bandIndex = clamp(int2(renderCoord * bandTransform.xy + bandTransform.zw), int2(0, 0), bandMax);

	#endif

	#if defined(SLUG_STROKE)

		int curveCount = (glyphData.w >> 9) & 0x07;
		if (curveCount == 0)
		{

	#endif

	#if defined(SLUG_LINEAR)

		if ((glyphData.w & 0x0100) == 0)
		{

	#endif

	#if defined(SLUG_MULTICOLOR)

		float4 finalColor = float4(0.0, 0.0, 0.0, 0.0);
		int2 colorLoc = glyphData.xy;

		int layerCount = int(TexelLoad2D(bandData, colorLoc).x);
		colorLoc.x += 1;

		for (int layer = 0; layer < layerCount; layer++)
		{
			int2 glyphLoc = int2(TexelLoad2D(bandData, int2(colorLoc.x, colorLoc.y)).xy);
			uint2 layerData = TexelLoad2D(bandData, int2(colorLoc.x + 1, colorLoc.y)).xy;
			colorLoc.x += 2;

	#else

		int2 glyphLoc = glyphData.xy;

	#endif

			float xcov = 0.0;
			float xwgt = 0.0;

			#if defined(SLUG_SUPER)

				renderCoord.y -= emsPerPixel.y * (float(sampleCount.y - 1) / float(sampleCount.y * 2));
				for (sampleIndex = 0; sampleIndex < sampleCount.y; sampleIndex++)
				{
					bandIndex.y = clamp(int(renderCoord.y * bandTransform.y + bandTransform.w), 0, bandMax.y);

			#endif

			uint4 hbandData = TexelLoad2D(bandData, int2(glyphLoc.x + bandIndex.y, glyphLoc.y));
			int2 hbandLoc = CalcBandLoc(glyphLoc, hbandData.y);

			for (curveIndex = 0; curveIndex < int(hbandData.x); curveIndex++)
			{
				int2 curveLoc = int2(TexelLoad2D(bandData, int2(hbandLoc.x + curveIndex, hbandLoc.y)).xy);
				float4 p12 = TexelLoad2D(curveData, curveLoc) - float4(renderCoord, renderCoord);
				float2 p3 = TexelLoad2D(curveData, int2(curveLoc.x + 1, curveLoc.y)).xy - renderCoord;

				if (max(max(p12.x, p12.z), p3.x) * pixelsPerEm.x < -0.5) break;

				slug_code code = CalcRootCode(p12.y, p12.w, p3.y);
				if (TestCurve(code))
				{
					float2 r = SolveHorizPoly(p12, p3) * pixelsPerEm.x;

					if (TestRoot1(code))
					{
						xcov += saturate(r.x + 0.5);
						xwgt = max(xwgt, saturate(1.0 - abs(r.x) * 2.0));
					}

					if (TestRoot2(code))
					{
						xcov -= saturate(r.y + 0.5);
						xwgt = max(xwgt, saturate(1.0 - abs(r.y) * 2.0));
					}
				}
			}

			#if defined(SLUG_SUPER)

					renderCoord.y += emsPerPixel.y / float(sampleCount.y);
				}

			#endif

			float ycov = 0.0;
			float ywgt = 0.0;

			#if defined(SLUG_SUPER)

				renderCoord = float2(pixelPosition.x - emsPerPixel.x * (float(sampleCount.x - 1) / float(sampleCount.x * 2)), pixelPosition.y);
				for (sampleIndex = 0; sampleIndex < sampleCount.x; sampleIndex++)
				{
					bandIndex.x = clamp(int(renderCoord.x * bandTransform.x + bandTransform.z), 0, bandMax.x);

			#endif

			uint4 vbandData = TexelLoad2D(bandData, int2(glyphLoc.x + bandMax.y + 1 + bandIndex.x, glyphLoc.y));
			int2 vbandLoc = CalcBandLoc(glyphLoc, vbandData.y);

			for (curveIndex = 0; curveIndex < int(vbandData.x); curveIndex++)
			{
				int2 curveLoc = int2(TexelLoad2D(bandData, int2(vbandLoc.x + curveIndex, vbandLoc.y)).xy);
				float4 p12 = TexelLoad2D(curveData, curveLoc) - float4(renderCoord, renderCoord);
				float2 p3 = TexelLoad2D(curveData, int2(curveLoc.x + 1, curveLoc.y)).xy - renderCoord;

				if (max(max(p12.y, p12.w), p3.y) * pixelsPerEm.y < -0.5) break;

				slug_code code = CalcRootCode(p12.x, p12.z, p3.x);
				if (TestCurve(code))
				{
					float2 r = SolveVertPoly(p12, p3) * pixelsPerEm.y;

					if (TestRoot1(code))
					{
						ycov -= saturate(r.x + 0.5);
						ywgt = max(ywgt, saturate(1.0 - abs(r.x) * 2.0));
					}

					if (TestRoot2(code))
					{
						ycov += saturate(r.y + 0.5);
						ywgt = max(ywgt, saturate(1.0 - abs(r.y) * 2.0));
					}
				}
			}

			#if defined(SLUG_SUPER)

					renderCoord.x += emsPerPixel.x / float(sampleCount.x);
				}

				xcov *= 1.0 / float(sampleCount.y);
				ycov *= 1.0 / float(sampleCount.x);

			#endif

			float coverage = CalcCoverage(xcov, ycov, xwgt, ywgt, glyphData.w);

	#if defined(SLUG_MULTICOLOR)

			finalColor = ApplyLayerColor(layerData, coverage, finalColor);

			#if defined(SLUG_SUPER)

				renderCoord.x = pixelPosition.x;

			#endif
		}

		return (float4(finalColor.xyz, finalColor.w));

	#else

		return (float4(vertexColor.xyz, coverage));

	#endif

	#if defined(SLUG_LINEAR)

		}
		else
		{
			#if defined(SLUG_MULTICOLOR)

				float4 finalColor = float4(0.0, 0.0, 0.0, 0.0);
				int2 colorLoc = glyphData.xy;

				int layerCount = int(TexelLoad2D(bandData, colorLoc).x);
				colorLoc.x += 1;

				for (int layer = 0; layer < layerCount; layer++)
				{
					int2 glyphLoc = int2(TexelLoad2D(bandData, int2(colorLoc.x, colorLoc.y)).xy);
					uint2 layerData = TexelLoad2D(bandData, int2(colorLoc.x + 1, colorLoc.y)).xy;
					colorLoc.x += 2;

			#else

				int2 glyphLoc = glyphData.xy;

			#endif

					float xcov = 0.0;
					float xwgt = 0.0;

					#if defined(SLUG_SUPER)

						renderCoord.y -= emsPerPixel.y * (float(sampleCount.y - 1) / float(sampleCount.y * 2));
						for (sampleIndex = 0; sampleIndex < sampleCount.y; sampleIndex++)
						{
							bandIndex.y = clamp(int(renderCoord.y * bandTransform.y + bandTransform.w), 0, bandMax.y);

					#endif

					uint4 hbandData = TexelLoad2D(bandData, int2(glyphLoc.x + bandIndex.y, glyphLoc.y));
					int2 hbandLoc = CalcBandLoc(glyphLoc, hbandData.y);

					for (curveIndex = 0; curveIndex < int(hbandData.x); curveIndex++)
					{
						int2 curveLoc = int2(TexelLoad2D(bandData, int2(hbandLoc.x + curveIndex, hbandLoc.y)).xy);
						float4 p12 = TexelLoad2D(curveData, curveLoc) - float4(renderCoord, renderCoord);

						if (max(p12.x, p12.z) * pixelsPerEm.x < -0.5) break;

						if ((asint(p12.y) ^ asint(p12.w)) < 0)
						{
							float r = (p12.x - p12.y * (p12.z - p12.x) / (p12.w - p12.y)) * pixelsPerEm.x;
							float s = saturate(r + 0.5);
							xcov += (p12.y < 0.0) ? -s : s;
							xwgt = max(xwgt, saturate(1.0 - abs(r) * 2.0));
						}
					}

					#if defined(SLUG_SUPER)

							renderCoord.y += emsPerPixel.y / float(sampleCount.y);
						}

					#endif

					float ycov = 0.0;
					float ywgt = 0.0;

					#if defined(SLUG_SUPER)

						renderCoord = float2(pixelPosition.x - emsPerPixel.x * (float(sampleCount.x - 1) / float(sampleCount.x * 2)), pixelPosition.y);
						for (sampleIndex = 0; sampleIndex < sampleCount.x; sampleIndex++)
						{
							bandIndex.x = clamp(int(renderCoord.x * bandTransform.x + bandTransform.z), 0, bandMax.x);

					#endif

					uint4 vbandData = TexelLoad2D(bandData, int2(glyphLoc.x + bandMax.y + 1 + bandIndex.x, glyphLoc.y));
					int2 vbandLoc = CalcBandLoc(glyphLoc, vbandData.y);

					for (curveIndex = 0; curveIndex < int(vbandData.x); curveIndex++)
					{
						int2 curveLoc = int2(TexelLoad2D(bandData, int2(vbandLoc.x + curveIndex, vbandLoc.y)).xy);
						float4 p12 = TexelLoad2D(curveData, curveLoc) - float4(renderCoord, renderCoord);

						if (max(p12.y, p12.w) * pixelsPerEm.y < -0.5) break;

						if ((asint(p12.x) ^ asint(p12.z)) < 0)
						{
							float r = (p12.y - p12.x * (p12.w - p12.y) / (p12.z - p12.x)) * pixelsPerEm.y;
							float s = saturate(r + 0.5);
							ycov += (p12.z < 0.0) ? -s : s;
							ywgt = max(ywgt, saturate(1.0 - abs(r) * 2.0));
						}
					}

					#if defined(SLUG_SUPER)

							renderCoord.x += emsPerPixel.x / float(sampleCount.x);
						}

						xcov *= 1.0 / float(sampleCount.y);
						ycov *= 1.0 / float(sampleCount.x);

					#endif

					float coverage = CalcCoverage(xcov, ycov, xwgt, ywgt, glyphData.w);

			#if defined(SLUG_MULTICOLOR)

					finalColor = ApplyLayerColor(layerData, coverage, finalColor);

					#if defined(SLUG_SUPER)

						renderCoord.x = pixelPosition.x;

					#endif
				}

				return (float4(finalColor.xyz, finalColor.w));

			#else

				return (float4(vertexColor.xyz, coverage));

			#endif
		}

	#endif

	#if defined(SLUG_STROKE)

		}
		else
		{
			curveCount += glyphData.x;

			float xcov = 0.0;
			float xwgt = 0.0;
			float4 p12 = TexelLoad2D(curveData, glyphData.xy) - float4(renderCoord, renderCoord);

			curveIndex = glyphData.x;
			do
			{
				float4 p3 = TexelLoad2D(curveData, int2(++curveIndex, glyphData.y)) - float4(renderCoord, renderCoord);

				slug_code code = CalcRootCode(p12.y, p12.w, p3.y);
				if (TestCurve(code))
				{
					float2 r = SolveHorizPoly(p12, p3.xy) * pixelsPerEm.x;

					if (TestRoot1(code))
					{
						xcov += saturate(r.x + 0.5);
						xwgt = max(xwgt, saturate(1.0 - abs(r.x) * 2.0));
					}

					if (TestRoot2(code))
					{
						xcov -= saturate(r.y + 0.5);
						xwgt = max(xwgt, saturate(1.0 - abs(r.y) * 2.0));
					}
				}

				p12 = p3;
			} while (curveIndex < curveCount);

			float ycov = 0.0;
			float ywgt = 0.0;
			p12 = TexelLoad2D(curveData, glyphData.xy) - float4(renderCoord, renderCoord);

			curveIndex = glyphData.x;
			do
			{
				float4 p3 = TexelLoad2D(curveData, int2(++curveIndex, glyphData.y)) - float4(renderCoord, renderCoord);

				slug_code code = CalcRootCode(p12.x, p12.z, p3.x);
				if (TestCurve(code))
				{
					float2 r = SolveVertPoly(p12, p3.xy) * pixelsPerEm.y;

					if (TestRoot1(code))
					{
						ycov -= saturate(r.x + 0.5);
						ywgt = max(ywgt, saturate(1.0 - abs(r.x) * 2.0));
					}

					if (TestRoot2(code))
					{
						ycov += saturate(r.y + 0.5);
						ywgt = max(ywgt, saturate(1.0 - abs(r.y) * 2.0));
					}
				}

				p12 = p3;
			} while (curveIndex < curveCount);

			float coverage = CalcCoverage(xcov, ycov, xwgt, ywgt, glyphData.w);
			return (float4(vertexColor.xyz, coverage));
		}

	#endif
}

float4 SlugRender(CurveSampler curveData, BandSampler bandData, float2 pixelPosition, float4 vertexColor, float4 bandTransform, int4 glyphData)
{
	float4 color = RenderGlyph(curveData, bandData, pixelPosition, vertexColor, bandTransform, glyphData);

	#if defined(SLUG_GRADIENT)

		int gradType = glyphData.w >> 13;
		if (gradType != 0)
		{
			int2 gradLoc = int2(TexelLoad2D(bandData, int2(glyphData.x - 1, glyphData.y)).xy);
			float3 gradGeom = TexelLoad2D(curveData, gradLoc).xyz;
			float4 gradColor1 = TexelLoad2D(curveData, int2(gradLoc.x + 1, gradLoc.y));
			float4 gradColor2 = TexelLoad2D(curveData, int2(gradLoc.x + 2, gradLoc.y));

			float t = (gradType == 1) ? saturate(gradGeom.x * pixelPosition.x + gradGeom.y * pixelPosition.y + gradGeom.z) : min(length(float2(pixelPosition.x * gradGeom.z - gradGeom.x, pixelPosition.y * gradGeom.z - gradGeom.y)), 1.0);
			color *= lerp(gradColor1, gradColor2, t);
		}

	#endif

	#if defined(SLUG_COVERAGE)

		return (float4(color.xyz * color.w, color.w * vertexColor.w));

	#elif defined(SLUG_INVERSE_COVERAGE)

		return (float4(color.xyz * (1.0 - color.w), color.w * vertexColor.w));

	#else

		return (float4(color.xyz, color.w * vertexColor.w));

	#endif
}

#if defined(VULKAN) || defined(GL_core_profile) || defined(GL_ES)

	#if defined(VULKAN)

		layout(location = 0) in vec4 vresult_color;
		layout(location = 1) in vec2 vresult_texcoord;
		layout(location = 2) flat in vec4 vresult_banding;
		layout(location = 3) flat in ivec4 vresult_glyph;

		layout(binding = 0) uniform sampler2D curveTexture;
		layout(binding = 1) uniform usampler2D bandTexture;

	#else

		in vec4 vresult_color;
		in vec2 vresult_texcoord;
		flat in vec4 vresult_banding;
		flat in ivec4 vresult_glyph;

		uniform sampler2D curveTexture;
		uniform usampler2D bandTexture;

	#endif

	layout(location = 0) out float4 fcolor;

	void main()
	{
		fcolor = SlugRender(curveTexture, bandTexture, vresult_texcoord, vresult_color, vresult_banding, vresult_glyph);
	}

#elif defined(__METAL_VERSION__)

	struct VertexStruct
	{
		float4 position [[position]];
		float4 color;
		float2 texcoord;
		float4 banding [[flat]];
		int4 glyph [[flat]];
	};

	fragment float4 PixelShader(VertexStruct vresult [[stage_in]], texture2d<float, access::read> curveTexture [[texture(0)]], texture2d<uint, access::read> bandTexture [[texture(1)]])
	{
		return (SlugRender(curveTexture, bandTexture, vresult.texcoord, vresult.color, vresult.banding, vresult.glyph));
	}

#elif defined(__PSSL__)

	struct VertexStruct
	{
		float4 color : U_COLOR;
		float2 texcoord : U_TEXCOORD;
		nointerp float4 banding : U_BANDING;
		nointerp int4 glyph : U_GLYPH;
	};

	struct TextureStruct
	{
		Texture2D curveTexture;
		Texture2D<uint4> bandTexture;
	};

	struct DataStruct
	{
		TextureStruct *textureData;
	};

	float4 main(VertexStruct vresult, DataStruct data : S_SRT_DATA) : S_TARGET_OUTPUT
	{
		return (SlugRender(data.textureData->curveTexture, data.textureData->bandTexture, vresult.texcoord, vresult.color, vresult.banding, vresult.glyph));
	}

#else	// HLSL

	struct VertexStruct
	{
		float4 position : SV_Position;
		float4 color : U_COLOR;
		float2 texcoord : U_TEXCOORD;
		nointerpolation float4 banding : U_BANDING;
		nointerpolation int4 glyph : U_GLYPH;
	};

	Texture2D curveTexture : register(t0);
	Texture2D<uint4> bandTexture : register(t1);

	float4 main(VertexStruct vresult) : SV_Target
	{
		return (SlugRender(curveTexture, bandTexture, vresult.texcoord, vresult.color, vresult.banding, vresult.glyph));
	}

#endif
